#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#ifndef AISL_WITHOUT_SSL
#include <openssl/err.h>
#endif

#include <aisl/aisl.h>
#include "stream.h"
#include "http.h"
#include "buffer.h"
#include "client.h"

#define FLAG_KEEPALIVE (1<<0)
#define FLAG_HANDSHAKE (1<<1)
#define FLAG_CAN_READ  (1<<2)
#define FLAG_CAN_WRITE (1<<3)

#define BUFFER_SIZE (16*1024)

struct aisl_client
{
  struct sockaddr_in   address;       /**< Client's address structure. */
  aisl_server_t        server;        /**< Server instance. */
  struct buffer        in;            /**< Client's input buffer. */
  struct buffer        out;           /**< Client's output buffer. */
  #ifndef AISL_WITHOUT_SSL
  SSL                * ssl;           /**< SSL pointer for HTTPS. */
  #endif
  time_t               timestamp;     /**< Last communication timestamp. */

  aisl_stream_t        stream;        /**< Pending client's stream. */
  int                  next_id;       /**< Stream id generator (even). */
  int                  flags;         /**< Client's flag bitmask. */
  int                  fd;            /**< Client's socket descriptor. */

  aisl_http_version_t  http_version;  /**< Client's http_version version. */
};


static bool
aisl_client_raise(aisl_client_t client, aisl_event_t event, ...)
{
  bool result;
  va_list vl_args;

  aisl_t instance = aisl_server_get_instance(client->server);

  va_start(vl_args, event);
  result = aisl_raise_vl(instance, client, event, vl_args);
  va_end(vl_args);

  return result;
}


static aisl_status_t
aisl_client_parse(aisl_client_t client, char * data, int32_t size)
{
  aisl_stream_t    s;
  http_parser_t    p = HTTP_PARSER_PENDING;

  int32_t bytes_left = size;

  switch (client->http_version)
  {
    case AISL_HTTP_1_0:
    case AISL_HTTP_1_1:

      s = client->stream;

      while ( p == HTTP_PARSER_PENDING )
      {

        switch ( aisl_stream_get_state(s) )
        {
          case AISL_STREAM_STATE_IDLE:
            p = http_10_parse_request(data, &size, client);
            break;

          case AISL_STREAM_STATE_WAIT_HEADER:
            p = http_10_parse_header(data, &size, client);
            break;

          case AISL_STREAM_STATE_WAIT_BODY:
            p = http_10_parse_body(data, &size, client);
            break;

          default: /* has input data, but request was already parsed */
            p = HTTP_PARSER_ERROR;
            continue;
        }
        // size now has number of parsed bytes
        data += size;
        bytes_left -= size;
        size = bytes_left;
      }

      if (p == HTTP_PARSER_FINISHED)
      {
        client->flags &= ~FLAG_CAN_READ;
        client->flags |= FLAG_CAN_WRITE;
      }

      break;

    case AISL_HTTP_2_0:

      break;
  }

  if (p == HTTP_PARSER_ERROR)
  {
    /* reply Bad Request here */
    aisl_client_close(client);

    return AISL_INPUT_ERROR;
  }


  if (size)
    buffer_shift(&client->in, client->in.length - size); /* reset buffer */

  return AISL_SUCCESS;
}


/* In HTTP 2.0 client->stream will be NULL if stream related data was completely
 * parsed. If it is not NULL, then stream expects additional data -> same like
 * in mono stream HTTP 1.0 
 */
static aisl_status_t
aisl_client_input(aisl_client_t client)
{
  int                l;

  char             * data = &client->in.data[ client->in.length ];
  int32_t            size = client->in.size - client->in.length;

  #ifndef AISL_WITHOUT_SSL
  if (client->ssl)
  {
    if ( !(client->flags & FLAG_HANDSHAKE) )
    {
      if ( (l = SSL_accept(client->ssl)) != 1 )
      {
        l = SSL_get_error(client->ssl, l);

        if (l == SSL_ERROR_WANT_READ || l == SSL_ERROR_WANT_WRITE)
          return AISL_IDLE;

        fprintf( stderr
               , "*AISL: SSL handshake fail: %s\n"
               , ERR_error_string(l, NULL) );

        aisl_client_close(client);
        return AISL_SYSCALL_ERROR;
      }

      client->flags |= FLAG_HANDSHAKE;
    }

    l = SSL_read(client->ssl, data, size) ;
  }
  else
  #endif
    l = recv( client->fd, data, size, 0);

  if (l > 0)
  {
    data = client->in.data;
    size = client->in.length + l;

    client->in.length = size;

    return aisl_client_parse(client, data, size);
  }
  else if (l<0)
  {
    if (client->ssl)
    {
      if (SSL_get_error(client->ssl, l) == SSL_ERROR_WANT_READ)
        return AISL_IDLE;
    }
    else
    {
      if(errno == EWOULDBLOCK)
        return AISL_IDLE;
    }
  }

  /* both: client disconnect + on read error  */
  /* todo: raise client error here */
  aisl_client_close(client);

  return AISL_SYSCALL_ERROR;
}


static aisl_status_t 
aisl_client_output(aisl_client_t client)
{
  int l;
  char * data;

  aisl_stream_t s = client->stream;

  /* while stream is not flushed, we should raise event */
  if( aisl_stream_get_chunked_output(s) )
  {
    /* in case of chunked output ( subscription for AISL_STREAM_OUTPUT event )
     * stream buffer will be initialized with OUTPUT_BUFFER_SIZE size, but
     * buffer->size will be used to carry amount of stored bytes
     * */
    l = aisl_stream_get_buffer_space(s);

    /*
    if (bsz < OUTPUT_BUFFER_SIZE)
    {
      if (buffer_clear(s->buffer, OUTPUT_BUFFER_SIZE) == 0)
        return false;

      s->buffer->size = bsz;
      bsz = OUTPUT_BUFFER_SIZE;
    }
    */

    if ( !(l < aisl_stream_get_buffer_size(s) / 2) )
      aisl_raise(
        aisl_server_get_instance(client->server),
        s,
        AISL_EVENT_STREAM_OUTPUT,
        l
      );
  }

  aisl_stream_get_buffer(s, &data, &l);

  if ( !l )
    return AISL_IDLE;

  l = (client->ssl) ?
        SSL_write(client->ssl, data, l) :
        send(     client->fd,  data, l, 0);

  if (l > 0)
  {
    aisl_stream_shift(s, l);

    /*
    if (s->state == STREAM_RESPONSE_READY && / * flushed * /
        s->buffer->size == 0) / * all sent * /
          */
    if ( ! aisl_stream_has_data(s) )
    {
      /* buffer_clear(s->buffer, 0); */

      /* data has been sent */

      if (client->flags & FLAG_KEEPALIVE)
      {
        aisl_stream_free(s);

        client->stream = aisl_stream_new(client, client->next_id++);
        if (client->stream != NULL )
          return AISL_SUCCESS;

        /* in case of malloc error it will not be error as long as request was
         * handled and we just close the connection.
         */
      }

      aisl_client_close(client);
    }

    return AISL_SUCCESS;
  }

  /* l < 0 */
  if (client->ssl)
  {
    if ( SSL_get_error(client->ssl, l) == SSL_ERROR_WANT_WRITE )
      return AISL_IDLE;
  }
  else
  {
    if (errno == EWOULDBLOCK)
      return AISL_IDLE;
  }

  aisl_client_close(client);

  return AISL_SYSCALL_ERROR;
}


#ifndef AISL_WITHOUT_SSL
aisl_client_t
aisl_client_new( aisl_server_t        server,
                 int                  fd,
                 struct sockaddr_in * addr,
                 SSL_CTX            * ssl_ctx )
#else
aisl_client_t
aisl_client_new( aisl_server_t        server,
                 int                  fd,
                 struct sockaddr_in * addr )
#endif
{
  aisl_client_t   client;
  aisl_stream_t   stream;

  if ( (client = calloc(1, sizeof(struct aisl_client))) != NULL )
  {
    memcpy(&client->address, addr, sizeof(struct sockaddr_in));
    client->server     = server;
    client->fd         = fd;
    client->next_id    = 2;
    client->http_version   = AISL_HTTP_1_0;
    client->timestamp  = time(NULL);
    client->flags      = FLAG_KEEPALIVE | FLAG_HANDSHAKE | FLAG_CAN_READ;

    if (buffer_init(&client->in, 2*BUFFER_SIZE) == 0)
    {
      memcpy(&client->out, &client->in, sizeof(struct buffer));

      stream = aisl_stream_new(client, 0);

      if (stream != NULL)
      {
        client->stream = stream;

        #ifdef AISL_WITHOUT_SSL

        return client;

        #else

        if ( !ssl_ctx )
          return client;

        if ((client->ssl = SSL_new(ssl_ctx)) != NULL )
        {
          SSL_set_fd(client->ssl, fd);
          return client;
        }

        #endif
      }
    }
    aisl_client_free(client);

  }

  return NULL;
}


void
aisl_client_free(aisl_client_t client)
{
  aisl_client_close(client);

  #ifndef AISL_WITHOUT_SSL
  if (client->ssl)
    SSL_free(client->ssl);
  #endif

  if (client->in.data)
    free(client->in.data);

  /* out buffer is a shared part of input buffer, so no need to free it */

  if (client->stream)
    aisl_stream_free(client->stream);

  free(client);
}


aisl_status_t
aisl_client_touch(aisl_client_t client)
{
  aisl_status_t result = AISL_IDLE,
                status = AISL_IDLE;

  /* input */
  if (client->flags & FLAG_CAN_READ)
  {
    if ( (result = aisl_client_input(client)) < 0 )
      return result;
  }

  /* output */
  if (client->flags & FLAG_CAN_WRITE)
  {
    if ( (status = aisl_client_output(client)) < 0 )
      return status;
  }


  /*
  if ((client->http_version==AISL_HTTP_2_0 || s->state<STREAM_REQUEST_READY) &&
      (client_input(client)) ) result = true;
  */
  /* output */
  /*
  s = list_index(client->streams, client->ostream);

  if ( s->flags & (STREAM_FLAG_OUTPUT_READY | STREAM_FLAG_OUTPUT_CHUNKED) )
    result = client_output(client);
  */
  /* update timestamp */

  if (result == AISL_IDLE)
    result = status;

  if (result == AISL_SUCCESS)
    client->timestamp = time(NULL);

  return result;
}


bool
aisl_client_is_timed_out(aisl_client_t client, uint32_t timeout)
{
  time_t now;

  time(&now);

  if ( !(now - client->timestamp < timeout) )
  {
    aisl_client_raise(client, AISL_EVENT_CLIENT_TIMEOUT);
    aisl_client_close(client);
    return true;
  }
  else
    return false;
}


int
aisl_client_get_socket(aisl_client_t client)
{
  return client->fd;
}


/* API Level ---------------------------------------------------------------- */

aisl_server_t
aisl_client_get_server(aisl_client_t client)
{
  return client->server;
}


bool
aisl_client_is_secure(aisl_client_t client)
{
  return (client->ssl == NULL) ? false : true;
}


bool
aisl_client_is_online(aisl_client_t client)
{
  return (client->fd == -1) ? false : true;
}


void
aisl_client_close(aisl_client_t client)
{
  if (client->fd != -1)
  {
    aisl_client_raise( client, AISL_EVENT_CLIENT_DISCONNECT );

    close(client->fd);
    shutdown(client->fd, SHUT_RDWR);
    client->fd = -1;
  }
}


aisl_http_version_t
aisl_client_get_http_version(aisl_client_t client)
{
  return client->http_version;
}
